﻿#region License Information
/* HeuristicLab
 * Copyright (C) 2002-2010 Heuristic and Evolutionary Algorithms Laboratory (HEAL)
 *
 * This file is part of HeuristicLab.
 *
 * HeuristicLab is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * HeuristicLab is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with HeuristicLab. If not, see <http://www.gnu.org/licenses/>.
 */
#endregion

using System;
using System.ServiceModel;
using HeuristicLab.Common;
using HeuristicLab.Tracing;
using HeuristicLab.Services.Hive.Common;
using System.ServiceModel.Description;

namespace HeuristicLab.Clients.Hive {
  public class ServiceClientFactory<T> {
    private static object locker = new object();

    private string endpointName;

    private string username;
    public string Username {
      get { return username; }
      set { 
        if(username != value) {
          username = value;
          this.factory = null;
        }
      }
    }

    private string password;
    public string Password {
      set {
        if (password != value) {
          password = value;
          this.factory = null;
        }
      }
    }

    private ChannelFactory<T> factory = null;
    private Disposable<T> disposableService = null;
    private int requestCount = 0;

    public ServiceClientFactory(string username, string password, string endpointName) {
      this.username = username;
      this.password = password;
      this.endpointName = endpointName;
    }

    private T CreateFacade(string endpointName) {
      lock (locker) {
        try {
          return CreateChannel(endpointName);
        }
        catch (EndpointNotFoundException ex) {
          OnExceptionOccured(ex);
        }
        return default(T);
      }
    }

    protected virtual T CreateChannel(string endpointName) {
      if (factory == null) {
        factory = new ChannelFactory<T>(endpointName);
        factory.Credentials.UserName.UserName = username;
        factory.Credentials.UserName.Password = password;
        factory.Credentials.ServiceCertificate.Authentication.CertificateValidationMode = System.ServiceModel.Security.X509CertificateValidationMode.None;
      }
      
      return factory.CreateChannel();
    }

    public Disposable<T> GetService() {
      lock (locker) {
        requestCount++;
        Logger.Debug("Request for ServiceProxy (count: " + requestCount + ")");
        if (disposableService != null) {
          if (GetServiceState() == CommunicationState.Faulted) {
            DisposeService();
          }
        }
        if (disposableService == null) {
          disposableService = new Disposable<T>(CreateFacade(this.endpointName));
          RegisterServiceEvents();
        }        
        return disposableService;
      }
    }

    private void RegisterServiceEvents() {
      disposableService.OnDisposing += new EventHandler(disposableService_OnDisposing);
      ((ICommunicationObject)disposableService.Obj).Faulted += new EventHandler(WcfServicePool_Faulted);
    }

    private void DeregisterServiceEvents() {
      disposableService.OnDisposing -= new EventHandler(disposableService_OnDisposing);
      ((ICommunicationObject)disposableService.Obj).Faulted -= new EventHandler(WcfServicePool_Faulted);
    }

    private CommunicationState GetServiceState() {
      return ((ICommunicationObject)disposableService.Obj).State;
    }

    void WcfServicePool_Faulted(object sender, EventArgs e) {
      OnExceptionOccured(new CommunicationException(e.ToString()));
    }

    private void disposableService_OnDisposing(object sender, EventArgs e) {
      DisposeService();
    }

    public void DisposeService() {
      lock (locker) {
        requestCount--;
        Logger.Debug("Disposing ServiceProxy (count: " + requestCount + ")");
        if (requestCount == 0) {
          try {
            DeregisterServiceEvents();
            ServiceUtil.DisposeWcfClient((ICommunicationObject)disposableService.Obj);
          }
          catch (Exception e) {
            OnExceptionOccured(e);
          }
          finally {
            disposableService = null;
          }
        } else if (requestCount < 0) {
          throw new ServiceClientFactoryException("requestCount cannot be less than 0.");
        }
      }
    }

    public event EventHandler<EventArgs<Exception>> ExceptionOccured;
    private void OnExceptionOccured(Exception e) {
      lock (locker) {
        var handler = ExceptionOccured;
        if (handler != null) handler(this, new EventArgs<Exception>(e));
      }
    }
  }

  public static class ServiceUtil {
    /// <summary>
    /// This method changes the endpoint-address while preserving the identity-certificate defined in the config file
    /// </summary>
    public static void ReplaceEndpointAddress(ServiceEndpoint endpoint, string hostAddress) {
      EndpointAddressBuilder builder = new EndpointAddressBuilder(endpoint.Address);
      UriBuilder uriBuilder = new UriBuilder(builder.Uri);
      uriBuilder.Host = hostAddress;
      builder.Uri = uriBuilder.Uri;
      endpoint.Address = builder.ToEndpointAddress();
    }

    /// <summary>
    /// Safely disposes a WCF client proxy object
    /// </summary>
    /// <param name="obj"></param>
    public static void DisposeWcfClient(ICommunicationObject obj) {
      if (obj != null) {
        if (obj.State != CommunicationState.Faulted &&
            obj.State != CommunicationState.Closed) {
          try { obj.Close(); }
          catch (CommunicationObjectFaultedException) { obj.Abort(); }
          catch (TimeoutException) { obj.Abort(); }
          catch (Exception e) {
            obj.Abort();
            Logger.Error(e);
          }
        } else
          obj.Abort();
      }
    }
  }
}
